#!/usr/local/bin/perl -w
#
# This Script takes directories out of a config file,
# tars them and copies the tared and zipped files over
# various locations configurable via an external config file.
# Call it with "./SGLBackup.pl --man" to get build-in manual.
#
# written and (c) by Sascha 'SieGeL' Pfalz <webmaster@saschapfalz.de>
#---------------------------------------------------------------------------
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#---------------------------------------------------------------------------
# $Id: SGLBackup.pl,v 1.16 2008/05/04 11:18:45 siegel Exp $
###################################################################################################

use strict;                         # Enforce all warnings
use Config::IniFiles;               # Used to parse the config file
use Getopt::Long;                   # Used for command-line parsing
use Pod::Usage;                     # Used for command-line help
use Time::HiRes qw(gettimeofday);   # For measurements (transfer speeds etc.)
use Net::FTP;                       # FTP Backup support
use File::Copy;                     # Filecopy/Move functions
use File::Basename;                 # Filename manipulation
use Data::Dumper;

###################################################################################################
# Prototypes
###################################################################################################

sub main();                         # Main entry of this script
sub ReadConfig($);                  # Sub to read and parse the configfile
sub getmicrotime;                   # Measures speed
sub TarAndZipSet($);                # Creates the backup file
sub ProcessBackup($$);              # Either filecopy or FTP the backup file and removes it from temp
sub UploadFTP($$$$$);               # Upload to a given FTP server
sub ExportOracle($);                # Allows to export an oracle schema or complete database
sub trim;                           # Trim function...
sub die_handler;                    # Catches DIE()
sub warn_handler;                   # Catches WARN()
sub WriteToLog($);                  # Logging function
sub ShowConfigAndExit;              # Displays parsed configuration
sub ExportMySQL($);                 # Exports MySQL Databases
sub CheckGenerations($$);           # Performs check on available generations for a given backupset
sub ReadFileCopies($$);             # Reads available backups from filesystem
sub ReadFTPCopies($);               # Reads available backups from FTP server
sub ReadMySQLDBList($$$$$$);        # Reads available MySQL databases and returns matching dbs

###################################################################################################
# Global Variables
###################################################################################################

use constant VER => '0.40';         # Version of this script
my %bd;                             # Backup Dirs, stored as DIR_0/NAME_0 pairs
my $cfgfile = 'config.ini';         # Config filename, can be changed via --config=<> parameter
my $cfg;                            # Configuration settings stored as hash ref
my $bsets   = '-1';                 # Which backupset to process. (-1 process all)
my $PATTERN_ACTIVE = 0;             # Global flag is set to 1 if pattern are found
use constant MODE_FCOPY   => 0;     # Mode FCOPY
use constant MODE_FTP     => 1;     # Mode FTP
use constant CALL_CMD     => 0;     # CheckGen is called via Commandline
use constant CALL_BAK     => 1;     # CheckGen is called after Backups
use constant FTP_TIMEOUT  => 30;    # FTP Connect timeout in seconds

###################################################################################################
# Start of script
###################################################################################################

$SIG{__WARN__}  = \&warn_handler;   # Install warn/die handlers
$SIG{__DIE__}   = \&die_handler;
my $mainst = getmicrotime;
$|=1;
main();
my $mainet = getmicrotime;
printf("Total time required: %2.3fs\n\n",abs($mainet-$mainst));
WriteToLog(sprintf("--- Everything done, total time required: %2.3fs ---",abs($mainet-$mainst)));
exit 0;

###################################################################################################
#    NAME: main()
# PURPOSE: Main entry of script, handles all processing
#   INPUT: None
#  RETURN: None
###################################################################################################

sub main()
  {
  my $help      = 0;  # Set to 1 if --help is choosen
  my $man       = 0;  # Set to 1 if --man is choosen
  my $showcfg   = 0;  # Set to 1 if --showconfig is choosen
  my $checkgen  = 0;  # Set to 1 if --checkgen is choosen
  my $backfile;

  printf("\nSGLBackup v%s written and (c) by Sascha 'SieGeL' Pfalz\n\n",VER);
  GetOptions ('help|?'          => \$help,
              'man'             => \$man,
              'config|f=s'      => \$cfgfile,
              'backupset|b=s'   => \$bsets,
              'showconfig|d'    => \$showcfg,
              'checkgen|cg'     => \$checkgen ) or pod2usage(2);
  pod2usage(1) if $help;
  pod2usage(-exitstatus => 0, -verbose => 2) if $man;
  $cfg = ReadConfig($cfgfile);
  if(!defined($cfg))
    {
    exit 10;
    }
  # Now make sure that in pattern matching mode the user did not specify
  # a specific set, as this won't work as expected:

  if($PATTERN_ACTIVE && ($bsets ne '-1' && $bsets ne ''))
    {
    print("You cannot specify specific backupsets when patterns are used!\nPlease refer to the documentation to learn why this is not allowed.\n\n");
    exit 10;
    }

  # Check if user wants to see the configuration list:

  if($showcfg)
    {
    ShowConfigAndExit();
    exit 0;
    }

  #
  # Test if user wants to  check available generations via commandline
  # Note that we perform this of course also after backups are made, this
  # is here to manually perform this only!
  #

  if($checkgen)
    {
    if($bsets eq '-1')
      {
      for(my $i = 0; $i < $cfg->{'maxsets'}; $i++)
        {
        CheckGenerations($i,CALL_CMD);
        }
      }
    else
      {
      my @params = split(/,/,$bsets);
      for(my $i = 0; $i < scalar @params; $i++)
        {
        if($params[$i] < 0 || $params[$i] > $cfg->{'maxsets'}-1)
          {
          printf(STDERR "ERROR: Set number %s out of range (Allowed is 0-%d)!\n\n",$params[$i],$cfg->{'maxsets'}-1);
          exit 10;
          }
        }
      for(my $i = 0; $i < scalar @params; $i++)
        {
        CheckGenerations($params[$i],CALL_CMD);
        }
      }
    print("\n");
    exit 0;
    }

  #
  # Perform the backup tasks, either complete or for selected backupsets
  #

  if($bsets eq '-1')
    {
    printf("Make backup of all %d sets, please wait.\n\n",$cfg->{'maxsets'});
    WriteToLog("--- Starting backup of all ".$cfg->{'maxsets'}." backupsets. ---");
    for(my $i = 0; $i < $cfg->{'maxsets'}; $i++)
      {
      if($cfg->{'dir'}[$i] ne '')
        {
        $backfile = TarAndZipSet($i);
        if(defined($backfile) && $backfile ne '')
          {
          ProcessBackup($backfile,$i);
          CheckGenerations($i,CALL_BAK);
          }
        }
      }
    }
  else
    {
    my @params = split(/,/,$bsets);
    for(my $i = 0; $i < scalar @params; $i++)
      {
      if($params[$i] < 0 || $params[$i] > $cfg->{'maxsets'}-1)
        {
        printf(STDERR "ERROR: Set number %s out of range (Allowed is 0-%d)!\n\n",$params[$i],$cfg->{'maxsets'}-1);
        exit 10;
        }
      }
    for(my $i = 0; $i < scalar @params; $i++)
      {
      $backfile = TarAndZipSet($params[$i]);
        if(defined($backfile) && $backfile ne '')
        {
        ProcessBackup($backfile,$params[$i]);
        CheckGenerations($params[$i],CALL_BAK);
        }
      }
    }
  }

###################################################################################################
#    NAME: ProcessBackup()
# PURPOSE: Processes on the backupped file according to settings for the passed setnr
#   INPUT: $bfile => The backupfile to work on
#          $setnr => Corresponding backupset number
#  RETURN: 0 if all was okay else an positive value to indicate an error condition
###################################################################################################

sub ProcessBackup($$)
  {
  my ($bfile,$setnr) = @_;
  my @dummy;

  # V0.38: If given setnr is a pattern, skip it.
  if($cfg->{'pattern'}[$setnr] == 1)
    {
    return(0);
    }
  @dummy = split(/,/,$cfg->{'mode'}[$setnr]);
  for(my $i = 0; $i < scalar @dummy; $i++)
    {
    if($dummy[$i] eq 'fcopy')
      {
      printf("Copying %s to %s (%d bytes)...",$bfile,$cfg->{'dest'}[$setnr],-s $bfile);
      WriteToLog(sprintf("Copying backup to %s (%d bytes)",$cfg->{'dest'}[$setnr],-s $bfile));
      my $rc = copy($bfile, $cfg->{'dest'}[$setnr]);
      if(!$rc)
        {
        printf(STDERR "ProcessBackup(copy) FAILED!\nError while copying backupfile: %s\n\n",$!);
        WriteToLog(sprintf("ERROR: FCOPY failed: %s (%s)",$!,$cfg->{'name'}[$setnr]));
        unlink($bfile);
        return(1);
        }
      print("done.\n");
      next;
      }
    if($dummy[$i] eq 'ftp')
      {
      $cfg->{'ftp'}[$setnr]=~/(.*)(\:)(.*)(@)(.*)/;
      my $ftpuser = $1;
      my $ftppass = $3;
      my $ftphost = $5;
      if($ftpuser eq '' || $ftppass eq '' || $ftphost eq '')
        {
        printf(STDERR "FTP: Logindata %s invalid, check settings!\n\n",$cfg->{'ftp'}[$setnr]);
        WriteToLog(sprintf("ERROR: FTP failed because config is wrong for set #%d (%s)",$setnr,$cfg->{'name'}[$setnr]));
        unlink($bfile);
        return(2);
        }
      printf("Sending file (%d bytes) to %s: ",-s $bfile, $ftphost);
      WriteToLog(sprintf("Sent backup via FTP to %s (%d bytes)",$ftphost,-s $bfile));
      if(!UploadFTP($ftpuser,$ftppass,$ftphost,$bfile,$setnr))
        {
        unlink($bfile);
        return(3);
        }
      print("done.\n");
      }
    }
  unlink($bfile);
  print("\n");
  WriteToLog("Backup set #".$setnr." successfully processed.");
  return(0);
  }

###################################################################################################
#    NAME: UploadFTP()
# PURPOSE: Uploads $bfile to given FTP server
#   INPUT: $fuser => FTP Username
#          $fpass => FTP Password
#          $fhost => FTP Host (May include the port in format HOST:PORT)
#          $bfile => Backupfile to sent to
#          $setnr => Number of corresponding backupset
#  RETURN: In case of an error 0 is returned else 1
###################################################################################################

sub UploadFTP($$$$$)
  {
  my ($fuser,$fpass,$fhost,$bfile,$setnr) = @_;
  my $ftp;
  my @hostdata = split(/\:/,$fhost);

  # V0.38: If given setnr is a pattern, skip it.
  if($cfg->{'pattern'}[$setnr] == 1)
    {
    return(0);
    }
  if(!defined($hostdata[1]))
    {
    $hostdata[1] = '21';  # Default port is 21
    }
  eval
    {
    $ftp = Net::FTP->new($hostdata[0], Debug => 0, Port => $hostdata[1], Timeout => FTP_TIMEOUT) or die "$@";
    };
  if($@)
    {
    $@=~/(.*)(at.*)/i;
    print(STDERR "UploadFTP(connect) FAILED: Reason: $1\n\n");
    WriteToLog(sprintf("ERROR: UploadFTP(#%d) FAILED: %s",$setnr,$1));
    return(0);
    }
  if(!$ftp->login($fuser,$fpass))
    {
    print(STDERR "UploadFTP(login) FAILED!\nFTP-Server says: ".$ftp->message."\n");
    WriteToLog(sprintf("ERROR: UploadFTP(#%d) LOGIN FAILED: %s",$setnr,$ftp->message));
    $ftp->quit;
    return(0);
    }
  if($cfg->{'ftpdir'}[$setnr] ne '')
    {
    if(!$ftp->cwd($cfg->{'ftpdir'}[$setnr]))
      {
      print(STDERR "UploadFTP(cwd) failed!\nFTP-Server says: ".$ftp->message."\n");
      WriteToLog(sprintf("ERROR: UploadFTP(#%d) CWD FAILED: %s",$setnr,$ftp->message));
      $ftp->quit;
      return(0);
      }
    }
  $ftp->binary;
  $ftp->put($bfile,basename($bfile));
  $ftp->quit;
  return(1);
  }

###################################################################################################
#    NAME: TarAndZipSet()
# PURPOSE: Creates the tar.gz file for passed set number
#   INPUT: Backup set number to work on
#  RETURN: Full pathname of created backupfile
#   NOTES: V0.2: Checks first if directory definition contains oracle export, in this
#                case we branch to the corresponding function and return the result of it.
###################################################################################################

sub TarAndZipSet($)
  {
  my ($setnr) = @_;
  my $targetname;

  # V0.38: If given setnr is a pattern, skip it.
  if($cfg->{'pattern'}[$setnr] == 1)
    {
    return(0);
    }
  printf("Processing set #%d (%s)...",$setnr,$cfg->{'name'}[$setnr]);
  WriteToLog(sprintf("Creating backup archive of set #%d (%s)",$setnr,$cfg->{'name'}[$setnr]));
  my $data = $cfg->{'dir'}[$setnr];
  if($data=~/^ORACLE\:/)
    {
    return(ExportOracle($setnr));
    }
  elsif($data=~/^MYSQL\:/)
    {
    return(ExportMySQL($setnr));
    }
  my $st = getmicrotime;
  if(!-d $cfg->{'dir'}[$setnr] || !-R $cfg->{'dir'}[$setnr])
    {
    printf(STDERR "TarAndZipSet() FAILED!\n'%s': %s!\n\n",$cfg->{'dir'}[$setnr],$!);
    WriteToLog(sprintf("ERROR: '%s': %s!",$cfg->{'dir'}[$setnr],$!));
    return(undef);
    }
  if(!-d $cfg->{'tmpdir'} || !-w $cfg->{'tmpdir'})
    {
    printf(STDERR "TarAndZipSet() FAILED!\nTempdir '%s' does not exist or is not writeable!\n\n",$cfg->{'tmpdir'});
    WriteToLog(sprintf("ERROR: Tempdir '%s' does not exist or is not writeable!",$cfg->{'tmpdir'}));
    return(undef);
    }
  # TAR the file:
  print(".");
  my @tiarray = localtime (time());
  my $myyear = $tiarray[5];
  if($myyear > 70)
	  {
	  $myyear+= 1900;
	  }
  else
	  {
	  $myyear+= 2000;
	  }
  $targetname = sprintf("%s/%s_%4d%02d%02d_%02d%02d%02d",$cfg->{'tmpdir'},$cfg->{'name'}[$setnr],$myyear,($tiarray[4]+1),$tiarray[3],$tiarray[2],$tiarray[1],$tiarray[0]);
  my $cmd = sprintf("%s %s %s.tar %s >/dev/null 2>>tar_stderr.log",$cfg->{'tar'},$cfg->{'taropts'},$targetname,$cfg->{'dir'}[$setnr]);
  system($cmd);
  if ($? == -1)
    {
    print(STDERR "TarAndZipSet(tar) FAILED!!!\nReason: $!\n\n");
    WriteToLog(sprintf("ERROR: TAR on set %s failed: %s",$cfg->{'name'}[$setnr],$!));
    return(undef);
    }
  elsif ($? & 127)
    {
    printf(STDERR "TarAndZipSet(tar) FAILED!!!\n%s died with signal %d, %s coredump\n\n",$cfg->{'tar'},($? & 127),  ($? & 128) ? 'with' : 'without');
    WriteToLog(sprintf("ERROR: TAR on set %s died with signal %d (%s coredump)",$cfg->{'name'}[$setnr],($? & 127),  ($? & 128) ? 'with' : 'without'));
    return(undef);
    }
  else
    {
    if(($? >> 8) > 0)
      {
      printf(STDERR "TarAndZipSet(tar) FAILED!!!\n%s exited with value %d\n\n",$cfg->{'tar'}, $? >> 8);
      WriteToLog(sprintf("ERROR: TAR on set %s exited with value %d",$cfg->{'name'}[$setnr],$? >> 8));
      unlink($targetname.'.tar');
      return(undef);
      }
    }
  #
  # And GZIP the file:
  #
  print(".");
  $cmd = sprintf("%s %s.tar >/dev/null 2>>packer_stderr.log",$cfg->{'gzip'},$targetname);
  system($cmd);
  if ($? == -1)
    {
    print(STDERR "TarAndZip(gzip) FAILED!!!\nReason: $!\n\n");
    WriteToLog(sprintf("ERROR: %s failed: %s",$cfg->{'gzip'},$!));
    return(undef);
    }
  elsif ($? & 127)
    {
    printf(STDERR "TarAndZip(gzip) FAILED!!!\n%s died with signal %d, %s coredump\n\n",$cfg->{'gzip'},($? & 127),  ($? & 128) ? 'with' : 'without');
    WriteToLog(sprintf("ERROR: %s died with signal %d (%s coredump)",$cfg->{'gzip'},($? & 127),  ($? & 128) ? 'with' : 'without'));
    return(undef);
    }
  else
    {
    if(($? >> 8) > 0)
      {
      printf(STDERR "TarAndZip(gzip) FAILED!!!\n%s exited with value %d\n\n",$cfg->{'gzip'}, $? >> 8);
      WriteToLog(sprintf("ERROR: %s exited with value %d",$cfg->{'gzip'},$? >> 8));
      unlink($targetname.'.tar'.$cfg->{'ext'});
      return(undef);
      }
    }
  my $et = getmicrotime;
  printf("done (%2.3fs).\n",abs($et-$st));
  WriteToLog(sprintf("Creation of Backup archive finished (%2.3fs)",abs($et-$st)));
  return($targetname.'.tar'.$cfg->{'ext'});
  }

###################################################################################################
#    NAME: ReadConfig()
# PURPOSE: Read and parse the configuration
#   INPUT: Configuration filename
#  RETURN: Hashref of parsed configuration
###################################################################################################

sub ReadConfig($)
  {
  my ($fname) = @_;
  my $cfg;
  my $config;
  my $i = 0;
  my $testparam;

  eval
    {
    $cfg = Config::IniFiles->new(-file => $fname, -nocase => 1) || die ( join("\n",  @Config::IniFiles::errors));
    $config->{'gzip'}         = trim($cfg->val('COMMAND','GZIP')) || die "Errors in $fname: No 'GZIP' tag defined!\n";
    $config->{'ext'}          = trim($cfg->val('COMMAND','GZIP_EXT')) || die "Errors in $fname: NO 'GZIP_EXT' tag defined!\n";
    $config->{'tar'}          = trim($cfg->val('COMMAND','TAR')) || die "Errors in $fname: No 'TAR' tag defined!\n";
    $config->{'taropts'}      = trim($cfg->val('COMMAND','TAROPTS'));
    $config->{'tmpdir'}       = trim($cfg->val('COMMAND','TMPDIR'));
    $config->{'logfile'}      = trim($cfg->val('COMMAND','LOGFILE'));
    $config->{'mysqldump'}    = trim($cfg->val('COMMAND','MYSQLDUMP'));
    $config->{'mysqldumpopts'}= trim($cfg->val('COMMAND','MYSQLDUMPOPTS'));
    $config->{'mknod'}        = trim($cfg->val('COMMAND','MKNOD'));
    while(1)
      {
      $config->{'dir'}[$i]    = trim($cfg->val('BACKUP_SETS','DIR_'.$i)) || last;
      $config->{'name'}[$i]   = trim($cfg->val('BACKUP_SETS','NAME_'.$i));
      $config->{'mode'}[$i]   = trim(lc($cfg->val('BACKUP_SETS','MODE_'.$i)));
      $config->{'maxgen'}[$i] = trim($cfg->val('BACKUP_SETS','MAXGEN_'.$i));
      $config->{'dest'}[$i]   = trim($cfg->val('BACKUP_SETS','DEST_'.$i));
      $config->{'ftp'}[$i]    = trim($cfg->val('BACKUP_SETS','FTP_'.$i));
      $config->{'ftpdir'}[$i] = trim($cfg->val('BACKUP_SETS','FTP_DEST_'.$i));
      $config->{'pattern'}[$i]= 0;
      $testparam              = trim($cfg->val('BACKUP_SETS','VLDB_'.$i));
      if(!defined($testparam) || $testparam eq '' || uc($testparam) ne 'YES')
        {
        $config->{'vldb'}[$i] = 0;
        }
      else
        {
        $config->{'vldb'}[$i] = 1;
        }
      my @test = split(/,/,$config->{'mode'}[$i]);
      for(my $a =0; $a < scalar @test; $a++)
        {
        if($test[$a] eq 'fcopy' && ( !defined($config->{'dest'}[$i]) || $config->{'dest'}[$i] eq ''))
          {
          die("CFG-ERROR: BackupSet #$i has FCOPY mode defined but no destination set!");
          }
        if($test[$a] eq 'ftp' && (!defined($config->{'ftp'}[$i]) || $config->{'ftp'}[$i] eq ''))
          {
          die("CFG-ERROR: BackupSet #$i has FTP mode defined but no FTP data given!!");
          }
        }
      $i++;
      undef($testparam);
      }
    };
  $config->{'maxsets'} = $i;
  if($@)
    {
    $@=~s/(.*)(at.*)/$1/g;
    chop($@);
    die (sprintf("%s\n\n",$@));
    }
  if($config->{'tmpdir'} eq '')
    {
    $config->{'tmpdir'} = '/tmp';
    }
  if(!-X $config->{'tar'})
    {
    die (sprintf("CFG: '%s' is not executable!!\n\n",$config->{'tar'}));
    }
  my @testfile = split(/ /,$config->{'gzip'});
  if(!-X $testfile[0])
    {
    die (sprintf("CFG: '%s' is not executable!!\n\n",$testfile[0]));
    }
  if(defined($config->{'mysqldump'}))
    {
    if(!-X $config->{'mysqldump'})
      {
      die(sprintf("CFG: '%s' not found / not executable!!!\n\n",$config->{'mysqldump'}));
      }
    }
  if(!defined($config->{'taropts'}) || $config->{'taropts'} eq '')
    {
    $config->{'taropts'} = '-chf';
    }
  if(!defined($config->{'mysqldumpopts'}))
    {
    $config->{'mysqldumpopts'} = '';
    }
  if(defined($config->{'mknod'}))
    {
    if(!-X $config->{'mknod'})
      {
      die(sprintf("CFG: '%s' not found / not executable!!!\n\n",$config->{'mknod'}));
      }
    }
  my $newmaxsets = $config->{'maxsets'};
  for(my $i = 0; $i < $config->{'maxsets'}; $i++)
    {
    if($config->{'dir'}[$i]=~/\*$/)
      {
      $PATTERN_ACTIVE = 1;
      $config->{'pattern'}[$i] = 1;       # Mark entry, so that we do not try to backup THIS entry
      if($config->{'dir'}[$i]=~/^MYSQL\:/)
        {
        my ($mysqldb,$mysqlhost,$mysqluser,$mysqlpass);
        my @dummy = split /\|/,$config->{'dir'}[$i];
        if(defined($dummy[1]) && $dummy[1] ne '')
          {
          $mysqldb = $dummy[1];
          }
        $dummy[0]=~s/^MYSQL://;
        my @dummy2 = split /\@/,$dummy[0];
        if(defined($dummy2[1]) && $dummy2[1] ne '')
          {
          $mysqlhost = $dummy2[1];
          }
        else
          {
          $mysqlhost = "localhost";
          }
        @dummy2 = split(/\//, $dummy2[0]);
        $mysqluser = $dummy2[0];
        if(!defined($mysqluser) || $mysqluser eq '')
          {
          printf(STDERR "ReadConfig() FAILED!!!\n\nCannot determine MySQL User name for backupset %s (Defined was %s) ????\n\n",$i,$config->{'dir'}[$i]);
          die ("Check configuration of listed backupset - Aborting\n");
          }
        $mysqlpass = (defined($dummy2[1]) ? $dummy2[1] : '');
        # Make pattern useable under perl (simple * is not enough)
        if($mysqldb eq '*')
          {
          $mysqldb = '.*';
          }
        else
          {
          $mysqldb=~s/\*/\.\*/;
          }
        my @dbs = ReadMySQLDBList($mysqluser,$mysqlpass,$mysqlhost,$mysqldb,$config->{'mysqldump'},$config->{'tmpdir'});
        # Now add the new found directories as new entries to our config array
        for(my $jobs = 0; $jobs < scalar @dbs; $jobs++)
          {
          $config->{'dir'}[$newmaxsets]     = sprintf("MYSQL:%s/%s@%s|%s",$mysqluser,$mysqlpass,$mysqlhost,$dbs[$jobs]);
          $config->{'name'}[$newmaxsets]    = $config->{'name'}[$i]."_".$dbs[$jobs];
          $config->{'mode'}[$newmaxsets]    = $config->{'mode'}[$i];
          $config->{'maxgen'}[$newmaxsets]  = $config->{'maxgen'}[$i];
          $config->{'dest'}[$newmaxsets]    = $config->{'dest'}[$i];
          $config->{'ftp'}[$newmaxsets]     = $config->{'ftp'}[$i];
          $config->{'ftpdir'}[$newmaxsets]  = $config->{'ftpdir'}[$i];
          $config->{'vldb'}[$newmaxsets]    = $config->{'vldb'}[$i];
          $config->{'pattern'}[$newmaxsets] = 0;
          $newmaxsets++;
          }
        }
      else
        {
        my @entries;
        my $targetdir = $config->{'dir'}[$i];
        my $pattern;
        $pattern   = substr($config->{'dir'}[$i],(rindex $config->{'dir'}[$i],"/")+1);
        $targetdir = substr($targetdir,0,rindex $targetdir,"/");
        # Avoid perl warning because * matches also NULL strings?! ...whatever that means ;)
        if($pattern eq '*')
          {
          $pattern = '.*';
          }
        else
          {
          $pattern=~s/\*/\.\*/;
          }
        opendir THISDIR, $targetdir || die "cannot open ".$config->{'dir'}[$i]." for reading pattern based directory!\n";
        my @files = readdir THISDIR;
        close THISDIR;
        @entries = grep !/^\.\.?$/,@files;              # Remove Dot files
        @entries = grep /^$pattern/,@entries;           # Remove everything else not matching our pattern
        @entries = grep -d "$targetdir/$_", @entries;   # Remove everything NOT a directory
        @entries = sort @entries;
        # Now add the new found directories as new entries to our config array
        for(my $jobs = 0; $jobs < scalar @entries; $jobs++)
          {
          $config->{'dir'}[$newmaxsets]     = $targetdir."/".$entries[$jobs];
          $config->{'name'}[$newmaxsets]    = $config->{'name'}[$i]."_".$entries[$jobs];
          $config->{'mode'}[$newmaxsets]    = $config->{'mode'}[$i];
          $config->{'maxgen'}[$newmaxsets]  = $config->{'maxgen'}[$i];
          $config->{'dest'}[$newmaxsets]    = $config->{'dest'}[$i];
          $config->{'ftp'}[$newmaxsets]     = $config->{'ftp'}[$i];
          $config->{'ftpdir'}[$newmaxsets]  = $config->{'ftpdir'}[$i];
          $config->{'vldb'}[$newmaxsets]    = $config->{'vldb'}[$i];
          $config->{'pattern'}[$newmaxsets] = 0;
          $newmaxsets++;
          }
        }
      }
    }
  $config->{'maxsets'} = $newmaxsets;
  # TO-DO: Check configured directories for accessability!
  return($config);
  }

###################################################################################################
#    NAME: getmicrotime
# PURPOSE: Determines seconds + microseconds. Used to measure processing speeds.
#   INPUT: None
#  RETURN: Float of current time in format sec.usec
###################################################################################################

sub getmicrotime
  {
  my ($sec,$usec) = gettimeofday;
  return(sprintf("%d.%d",$sec,$usec));
  }

###################################################################################################
#    NAME: ExportOracle()
# PURPOSE: Exports an Oracle Database with the help of exp
#   INPUT: SetNr -> Backupset to work on
#  RETURN: Full backupfile name with extension or undef in case of an error
###################################################################################################

sub ExportOracle($)
  {
  my ($setnr) = @_;

  my $st = getmicrotime;
  my $schema = "";
  my $orahost = "";
  my $data = $cfg->{'dir'}[$setnr];

  if(!defined($ENV{'ORACLE_HOME'}))
    {
    print(STDERR "ExportOracle() FAILED!\n\nERROR: ORACLE_HOME environment variable not found, Oracle Support disabled!\n\n");
    WriteToLog(sprintf("ERROR: ORACLE_HOME env var not found (%s)",$cfg->{'name'}[$setnr]));
    return(undef);
    }
  my @dummy = split /\|/,$data;
  if(defined($dummy[1]) && $dummy[1] ne '')
    {
    $schema = $dummy[1];
    }
  my $oralogin = $dummy[0];
  $oralogin=~s/^ORACLE://;
  my @dummy2 = split /\@/,$oralogin;
  if(defined($dummy2[1]) && $dummy2[1] ne '')
    {
    $orahost = "database ".$dummy2[1];
    }
  else
    {
    $orahost = "local database (".(defined($ENV{'ORACLE_SID'}) ? $ENV{'ORACLE_SID'} : 'No ORACLE_SID set ?!').")";
    }
  my $oraexport = sprintf("%s/bin/exp",$ENV{'ORACLE_HOME'});
  if(!-x $oraexport)
    {
    print(STDERR "ExportOracle() FAILED!\n\nERROR: Oracle Export not found / not executable!\n\n");
    WriteToLog(sprintf("ERROR: Oracle Export not found / not executable! (%s)",$cfg->{'name'}[$setnr]));
    return(undef);
    }
  if($cfg->{'vldb'}[$setnr])
    {
    return(ExportOracleVLDB($setnr,$oralogin,$schema,$orahost,$oraexport));
    }
  my $cmdline;
  my $targetname = "";
  my $targetlog  = "";
  print(".");
  my @tiarray = localtime (time());
  my $myyear = $tiarray[5];
  if($myyear > 70)
	  {
	  $myyear+= 1900;
	  }
  else
	  {
	  $myyear+= 2000;
	  }
  $targetname = sprintf("%s/%s_%4d%02d%02d_%02d%02d%02d.dmp",$cfg->{'tmpdir'},$cfg->{'name'}[$setnr],$myyear,($tiarray[4]+1),$tiarray[3],$tiarray[2],$tiarray[1],$tiarray[0]);
  $targetlog  = sprintf("%s/%s_%4d%02d%02d_%02d%02d%02d.log",$cfg->{'tmpdir'},$cfg->{'name'}[$setnr],$myyear,($tiarray[4]+1),$tiarray[3],$tiarray[2],$tiarray[1],$tiarray[0]);
  if($schema eq '')
    {
    WriteToLog(sprintf("Starting Oracle Export of whole %s",$orahost));
    $cmdline = sprintf("%s USERID=%s FILE=%s LOG=%s FULL=Y ROWS=Y INDEXES=Y CONSISTENT=Y STATISTICS=NONE >/dev/null 2>&1",$oraexport,$oralogin,$targetname,$targetlog);
    }
  else
    {
    WriteToLog(sprintf("Starting Oracle Export of schema %s from %s",$schema,$orahost));
    $cmdline = sprintf("%s USERID=%s FILE=%s LOG=%s OWNER=%s ROWS=Y INDEXES=Y CONSISTENT=Y STATISTICS=NONE >/dev/null 2>&1",$oraexport,$oralogin,$targetname,$targetlog,$schema);
    }
  system($cmdline);
  if ($? == -1)
    {
    print(STDERR "ExportOracle(exp) FAILED!!!\nReason: $!\n\n");
    WriteToLog(sprintf("ERROR: Oracle Export failed: %s (%s)",$!,$cfg->{'name'}[$setnr]));
    unlink($targetname);
    unlink($targetlog);
    return(undef);
    }
  elsif ($? & 127)
    {
    printf(STDERR "ExportOracle(exp) FAILED!!!\nexp died with signal %d, %s coredump\n\n",($? & 127),  ($? & 128) ? 'with' : 'without');
    WriteToLog(sprintf("ERROR: Oracle Export died with signal %d, %s coredump (%s)",($? & 127),  ($? & 128) ? 'with' : 'without',$cfg->{'name'}[$setnr]));
    unlink($targetname);
    unlink($targetlog);
    return(undef);
    }
  else
    {
    if(($? >> 8) > 0)
      {
      printf(STDERR "ExportOracle(exp) FAILED!!!\nexp exited with value %d\n\n", $? >> 8);
      WriteToLog(sprintf("ERROR: Oracle Export exited with value %d (%s)",$? >> 8,$cfg->{'name'}[$setnr]));
      unlink($targetname);
      unlink($targetlog);
      return(undef);
      }
    }
  # And TAR/GZIP the file:
  print(".");
  my $tarname = sprintf("%s/%s_%4d%02d%02d_%02d%02d%02d",$cfg->{'tmpdir'},$cfg->{'name'}[$setnr],$myyear,($tiarray[4]+1),$tiarray[3],$tiarray[2],$tiarray[1],$tiarray[0]);
  my $cmd     = sprintf("%s %s %s.tar %s %s >/dev/null 2>&1",$cfg->{'tar'},$cfg->{'taropts'},$tarname,$targetname,$targetlog);
  system($cmd);
  if ($? == -1)
    {
    print(STDERR "ExportOracle(tar) FAILED!!!\nReason: $!\n\n");
    WriteToLog(sprintf("ERROR: OraExp->Tar failed: %s (%s)",$!,$cfg->{'name'}[$setnr]));
    unlink($targetname);
    unlink($targetlog);
    return(undef);
    }
  elsif ($? & 127)
    {
    printf(STDERR "ExportOracle(tar) FAILED!!!\n%s died with signal %d, %s coredump\n\n",$cfg->{'tar'},($? & 127),  ($? & 128) ? 'with' : 'without');
    WriteToLog(sprintf("ERROR: OraExp->Tar died with signal %d, %s coredump (%s)",($? & 127),  ($? & 128) ? 'with' : 'without',$cfg->{'name'}[$setnr]));
    unlink($targetname);
    unlink($targetlog);
    return(undef);
    }
  else
    {
    if(($? >> 8) > 0)
      {
      printf(STDERR "ExportOracle(tar) FAILED!!!\n%s exited with value %d\n\n",$cfg->{'tar'}, $? >> 8);
      WriteToLog(sprintf("ERROR: OraExp->Tar exited with value %d (%s)",$? >> 8,$cfg->{'name'}[$setnr]));
      unlink($tarname.'.tar');
      unlink($targetname);
      unlink($targetlog);
      return(undef);
      }
    }
  # And GZIP the file (removing first the original files to save space):
  print(".");
  unlink($targetname);
  unlink($targetlog);
  $cmd = sprintf("%s %s.tar >/dev/null 2>&1",$cfg->{'gzip'},$tarname);
  system($cmd);
  if ($? == -1)
    {
    print(STDERR "ExportOracle(gzip) FAILED!!!\nReason: $!\n\n");
    WriteToLog(sprintf("ERROR: OraExp->%s failed: %s (%s)",$cfg->{'gzip'},$!,$cfg->{'name'}[$setnr]));
    unlink($tarname.'.tar');
    return(undef);
    }
  elsif ($? & 127)
    {
    printf(STDERR "ExportOracle(gzip) FAILED!!!\n%s died with signal %d, %s coredump\n\n",$cfg->{'gzip'},($? & 127),  ($? & 128) ? 'with' : 'without');
    WriteToLog(sprintf("ERROR: OraExp->%s died with signal %d, %s coredump (%s)",$cfg->{'gzip'},($? & 127),  ($? & 128) ? 'with' : 'without',$cfg->{'name'}[$setnr]));
    unlink($tarname.'.tar');
    return(undef);
    }
  else
    {
    if(($? >> 8) > 0)
      {
      printf(STDERR "ExportOracle(gzip) FAILED!!!\n%s exited with value %d\n\n",$cfg->{'gzip'}, $? >> 8);
      WriteToLog(sprintf("ERROR: OraExp->%s exited with value %d (%s)",$cfg->{'gzip'},$? >> 8,$cfg->{'name'}[$setnr]));
      unlink($tarname.'.tar'.$cfg->{'ext'});
      unlink($tarname.'.tar');
      return(undef);
      }
    }
  my $et = getmicrotime;
  printf("done (%2.3fs).\n",abs($et-$st));
  WriteToLog(sprintf("Oracle Export finished after %2.3fs",abs($et-$st)));
  return($tarname.'.tar'.$cfg->{'ext'});
  }

###################################################################################################
#    NAME: ExportOracleVLDB()
# PURPOSE: Exports an Oracle Database with the help of exp via a named pipe
#   INPUT: SetNr    -> Backupset to work on
#          Oralogin -> User/password@TNS
#          OraHost  -> Schema
#  RETURN: Full backupfile name with extension or undef in case of an error
###################################################################################################

sub ExportOracleVLDB($$$$$)
  {
  my ($setnr,$ologin,$schema,$orahost,$oraexport) = @_;
  my $st = getmicrotime;

  # Make sure that we have ONLY (!) FCOPY mode active for this backupset, FTP via named pipe would be useless:

  my @dummy = split(/,/,$cfg->{'mode'}[$setnr]);
  for(my $i = 0; $i < scalar @dummy; $i++)
    {
    if($dummy[$i] ne 'fcopy')
      {
      printf(STDERR "Only FCOPY mode is allowed for VLDB backup mode (Set #%d) - Skipping entry!",$setnr);
      WriteToLog(sprintf("ERROR: Only FCOPY mode is allowed for VLDB backup mode (Set #%d) - Skipping entry!",$setnr));
      }
    }
  # First we create the named pipe based on the name of the backupset:
  my $pipename = sprintf("exportpipe_%s.%d",$cfg->{'name'}[$setnr],$setnr);
  my $cmd = sprintf("%s %s p",$cfg->{'mknod'},$pipename);
  system($cmd);
  if ($? == -1)
    {
    print(STDERR "ExportOracleVLDB(mknod) FAILED!!!\nReason: $!\n\n");
    WriteToLog(sprintf("ERROR: OraExpVLDB(%s)->%s failed: %s (%s)",$cfg->{'name'}[$setnr],$cfg->{'mknod'},$!));
    return(undef);
    }
  elsif($? & 127)
    {
    printf(STDERR "ExportOracleVLDB(mknod) FAILED!!!\n%s died with signal %d, %s coredump\n\n",$cfg->{'mknod'},($? & 127),  ($? & 128) ? 'with' : 'without');
    WriteToLog(sprintf("ERROR: OraExpVLDB->%s died with signal %d, %s coredump (%s)",$cfg->{'mknod'},($? & 127),  ($? & 128) ? 'with' : 'without',$cfg->{'name'}[$setnr]));
    return(undef);
    }
  else
    {
    if(($? >> 8) > 0)
      {
      printf(STDERR "ExportOracleVLDB(mknod) FAILED!!!\n%s exited with value %d (%s)\n\n",$cfg->{'mknod'}, $? >> 8,$!);
      WriteToLog(sprintf("ERROR: OraExpVLDB(%s)->%s exited with value %d",$cfg->{'name'}[$setnr],$cfg->{'mknod'},$? >> 8));
      return(undef);
      }
    }
  # Now we have to start the packer in background, as the pipe is FIFO, therefor first the writer process:
  my $targetname = "";
  my $targetlog  = "";
  print(".");
  my @tiarray = localtime (time());
  my $myyear = $tiarray[5];
  if($myyear > 70)
	  {
	  $myyear+= 1900;
	  }
  else
	  {
	  $myyear+= 2000;
	  }
  $targetname = sprintf("%s/%s_%4d%02d%02d_%02d%02d%02d.dmp%s",$cfg->{'dest'}[$setnr],$cfg->{'name'}[$setnr],$myyear,($tiarray[4]+1),$tiarray[3],$tiarray[2],$tiarray[1],$tiarray[0],$cfg->{'ext'});
  $cmd = sprintf("%s --stdout < %s > %s &",$cfg->{'gzip'},$pipename,$targetname);
  system($cmd);
  if ($? == -1)
    {
    print(STDERR "ExportOracleVLDB(gzip_in_bg) FAILED!!!\nReason: $!\n\n");
    WriteToLog(sprintf("ERROR: OraExpVLDB(%s)->%s failed: %s (%s)",$cfg->{'name'}[$setnr],$cfg->{'gzip'},$!));
    unlink $pipename;
    return(undef);
    }
  elsif($? & 127)
    {
    printf(STDERR "ExportOracleVLDB(gzip_in_bg) FAILED!!!\n%s died with signal %d, %s coredump\n\n",$cfg->{'gzip'},($? & 127),  ($? & 128) ? 'with' : 'without');
    WriteToLog(sprintf("ERROR: OraExpVLDB(%s)->%s died with signal %d, %s coredump",$cfg->{'name'}[$setnr],$cfg->{'gzip'},($? & 127),  ($? & 128) ? 'with' : 'without'));
    unlink $pipename;
    return(undef);
    }
  else
    {
    if(($? >> 8) > 0)
      {
      printf(STDERR "ExportOracleVLDB(gzip_in_bg) FAILED!!!\n%s exited with value %d (%s)\n\n",$cfg->{'gzip'}, $? >> 8,$!);
      WriteToLog(sprintf("ERROR: OraExpVLDB(%s)->%s exited with value %d",$cfg->{'name'}[$setnr],$cfg->{'gzip'},$? >> 8));
      unlink $pipename;
      return(undef);
      }
    }
  # Packer is now started, next we must start export and pointing the dump file to our pipe:
  if($schema eq '')
    {
    WriteToLog(sprintf("Starting Oracle Export of whole %s",$orahost));
    $cmd = sprintf("%s USERID=%s FILE=%s FULL=Y ROWS=Y INDEXES=Y CONSISTENT=Y STATISTICS=NONE >/dev/null 2>&1",$oraexport,$ologin,$pipename);
    }
  else
    {
    WriteToLog(sprintf("Starting Oracle Export of schema %s from %s",$schema,$orahost));
    $cmd = sprintf("%s USERID=%s FILE=%s OWNER=%s ROWS=Y INDEXES=Y CONSISTENT=Y STATISTICS=NONE >/dev/null 2>&1",$oraexport,$ologin,$pipename,$schema);
    }
  system($cmd);
  if ($? == -1)
    {
    print(STDERR "ExportOracleVLDB(exp) FAILED!!!\nReason: $!\n\n");
    WriteToLog(sprintf("ERROR: ExportOracleVLDB failed: %s (%s)",$!,$cfg->{'name'}[$setnr]));
    unlink($targetname);
    unlink($pipename);
    return(undef);
    }
  elsif ($? & 127)
    {
    printf(STDERR "ExportOracleVLDB(exp) FAILED!!!\nexp died with signal %d, %s coredump\n\n",($? & 127),  ($? & 128) ? 'with' : 'without');
    WriteToLog(sprintf("ERROR: ExportOracleVLDB died with signal %d, %s coredump (%s)",($? & 127),  ($? & 128) ? 'with' : 'without',$cfg->{'name'}[$setnr]));
    unlink($targetname);
    unlink($pipename);
    return(undef);
    }
  else
    {
    if(($? >> 8) > 0)
      {
      printf(STDERR "ExportOracleVLDB(exp) FAILED!!!\nexp exited with value %d\n\n", $? >> 8);
      WriteToLog(sprintf("ERROR: ExportOracleVLDB exited with value %d (%s)",$? >> 8,$cfg->{'name'}[$setnr]));
      unlink($targetname);
      unlink($pipename);
      return(undef);
      }
    }
  # Finally we have to remove the named pipe:
  unlink $pipename;
  my $et = getmicrotime;
  printf("done (%2.3fs).\n",abs($et-$st));
  # As this is a special behavour here, we call CheckGenerations() directly from here and return
  # a NULL name to our callee functions, avoiding any further action on that backupset.
  CheckGenerations($setnr,CALL_BAK);
  return(undef);
  }

###################################################################################################
#    NAME: trim
# PURPOSE: Simple trim() function...unbelievable that Perl does not supply such a trivial function...
#   INPUT: String or array from where to trim leading and trailing spaces
#  RETURN: Trimmed version of passed string or array
###################################################################################################

sub trim
  {
  my @out = @_;
  if(@_)
    {
    eval
      {
      for (@out)
        {
        if(defined($_))
          {
          s/^\s+//;
          s/\s+$//;
          }
        }
      };
    }
  return wantarray ? @out : $out[0];
  }

###################################################################################################
#    NAME: warn_handler
# PURPOSE: Signal handler for warn() calls (routes to WriteToLog()
#   INPUT: None
#  RETURN: None
###################################################################################################

sub warn_handler
  {
  my $warntxt;
  my $showerr;

  $warntxt = join(', ',@_);
  $warntxt=~s/\n//g;
  $warntxt=~/(.*\s+)(at\s+)(.*)(\s+)(line\s+\d{0,}).*/;

  if(defined($3) && $3 ne '')
    {
    my $appname = basename($3);
    my $errline = $5;
    $showerr = sprintf("[%s: %s] %s",$appname,$errline,$1);
    }
  else
    {
    $showerr = $warntxt;
    }
  print STDERR $showerr."\n";
  WriteToLog($showerr);
  }

###################################################################################################
#    NAME: die_handler
# PURPOSE: Signal handler for die() calls (adds date/time/file/linenumber):
#   INPUT: None
#  RETURN: None
#   NOTES: Catches die(), writes message to logfile and exits the script!
###################################################################################################

sub die_handler
  {
  my ($error) = @_;

  $error=~s/\n//g;
  $error=~/(.*\s+)(at\s+)(.*)(\s+)(line\s+\d{0,}).*/;

  if(defined($3) && $3 ne '')
    {
    my $appname = basename($3);
    my $errline = $5;
    my $showerr = sprintf("[%s: %s] %s",$appname,$errline,$1);
    WriteToLog($showerr);
    }
  die $error."\n\n";
  exit;
  }

###################################################################################################
#    NAME: WriteToLog()
# PURPOSE: General logging function
#   INPUT: Logstring to write
#  RETURN: None
###################################################################################################

sub WriteToLog($)
  {
  my ($logstring) = @_;

  my @tiarray = localtime (time());

  # If no logfile is defined we skip writing to log:
  if(!defined($cfg->{'logfile'}) || $cfg->{'logfile'} eq "")
    {
    return;
    }
  if($tiarray[5] > 70)
    {
    $tiarray[5]+=1900;
    }
  else
    {
    $tiarray[5]+=2000;
    }
  open(FH,">>".$cfg->{'logfile'}) || die "Unable to write to $cfg->{'logfile'} ?!";
  if(!defined($logstring)) { $logstring = ''; }
  my $buffer = sprintf("[%02d/%02d/%4d %02d:%02d:%02d]: %s\n",($tiarray[4]+1),$tiarray[3],$tiarray[5],$tiarray[2],$tiarray[1],$tiarray[0],$logstring);
  print FH $buffer;
  close(FH);
  }

###################################################################################################
#    NAME: ShowConfigAndExit
# PURPOSE: Lists the parsed configuration. Useful to check if everything is correctly
#          configured before using SGLBackup via cronjob. (V0.33+)
#   INPUT: None
#  RETURN: None
###################################################################################################

sub ShowConfigAndExit
  {
  my $target;

  print("Show configuration:\n\n");
  printf("TAR program (Options)....: %s %s\n",$cfg->{'tar'},$cfg->{'taropts'});
  printf("Packer program...........: %s\n",$cfg->{'gzip'});
  printf("Packer extension.........: %s\n",$cfg->{'ext'});
  printf("Temporary directory......: %s\n",$cfg->{'tmpdir'});
  printf("Logfile to write.........: %s\n",(defined($cfg->{'logfile'})      ? $cfg->{'logfile'} : 'No logfile configured'));
  printf("ORACLE_HOME available....: %s\n",(defined($ENV{'ORACLE_HOME'})    ? 'Yes' : 'No'));
  printf("mysqldump utility........: %s\n",(defined($cfg->{'mysqldump'})    ? $cfg->{'mysqldump'} : 'Not available'));
  printf("mysqldump options........: %s\n",$cfg->{'mysqldumpopts'});
  printf("mknod utility............: %s\n",(defined($cfg->{'mknod'})        ? $cfg->{'mknod'} : 'Not available'));
  printf("Number of backupsets.....: %d\n",$cfg->{'maxsets'});
  print("\nConfigured backup sets:\n\n");

  for(my $i = 0; $i < $cfg->{'maxsets'}; $i++)
    {
    printf("Backupset number (name)..: %d (%s)\n",$i, $cfg->{'name'}[$i]);
    printf("Backup source............: %s\n",$cfg->{'dir'}[$i]);
    $target = "";
    my @test = split(/,/,$cfg->{'mode'}[$i]);
    for(my $a =0; $a < scalar @test; $a++)
      {
      if($test[$a] eq 'fcopy')
        {
        $target = "FCOPY: ".$cfg->{'dest'}[$i];
        }
      if($test[$a] eq 'ftp')
        {
        $target = "FTP: ".$cfg->{'ftp'}[$i]." (".$cfg->{'ftpdir'}[$i].")";
        }
      printf("Target destination.......: %s\n",$target);
      print ("Defines pattern match....: ");
      if($cfg->{'pattern'}[$i] == 1)
        {
        print("Yes\n");
        }
      else
        {
        print("No\n");
        }
      print ("Backup using named pipe..: ");
      if($cfg->{'vldb'}[$i] == 1)
        {
        print("Yes\n");
        }
      else
        {
        print("No\n");
        }
      }
    print("\n");
    }
  }

###################################################################################################
#    NAME: ExportMySQL()
# PURPOSE: Exports an MySQL Database with the help of mysqldump
#   INPUT: SetNr -> Backupset to work on
#  RETURN: Full backupfile name with extension or undef in case of an error
#   NOTES: Format of connect string is <username>/<password>@<hostname>|<database>
###################################################################################################

sub ExportMySQL($)
  {
  my ($setnr)   = @_;
  my $st        = getmicrotime;
  my $mysqluser = "";
  my $mysqlpass = "";
  my $mysqldb   = "";
  my $mysqlhost = "";
  my $data      = $cfg->{'dir'}[$setnr];
  my $cmdline;
  my $targetname = "";

  my @dummy = split /\|/,$data;
  if(defined($dummy[1]) && $dummy[1] ne '')
    {
    $mysqldb = $dummy[1];
    }
  $dummy[0]=~s/^MYSQL://;
  my @dummy2 = split /\@/,$dummy[0];
  if(defined($dummy2[1]) && $dummy2[1] ne '')
    {
    $mysqlhost = $dummy2[1];
    }
  else
    {
    $mysqlhost = "localhost";
    }
  @dummy2 = split(/\//, $dummy2[0]);
  $mysqluser = $dummy2[0];
  if(!defined($mysqluser) || $mysqluser eq '')
    {
    printf(STDERR "ExportMySQL() FAILED!!!\nCannot determine MySQL User name ????\n\n");
    return(undef);
    }
  $mysqlpass = (defined($dummy2[1]) ? $dummy2[1] : '');
  print(".");
  my @tiarray = localtime (time());
  my $myyear = $tiarray[5];
  if($myyear > 70)
	  {
	  $myyear+= 1900;
	  }
  else
	  {
	  $myyear+= 2000;
	  }
  $targetname = sprintf("%s/%s_%4d%02d%02d_%02d%02d%02d.sql",$cfg->{'tmpdir'},$cfg->{'name'}[$setnr],$myyear,($tiarray[4]+1),$tiarray[3],$tiarray[2],$tiarray[1],$tiarray[0]);
  if($mysqldb eq '')
    {
    WriteToLog(sprintf("Starting MySQL Export of whole database from host %s",$mysqlhost));
    $cmdline = sprintf("%s --user=%s --password=%s --host=%s --all-databases %s >%s",$cfg->{'mysqldump'},$mysqluser,$mysqlpass,$mysqlhost,$cfg->{'mysqldumpopts'},$targetname);
    }
  else
    {
    WriteToLog(sprintf("Starting MySQL Export of database %s from host %s",$mysqldb,$mysqlhost));
    $cmdline = sprintf("%s --user=%s --password=%s --host=%s %s %s >%s",$cfg->{'mysqldump'},$mysqluser,$mysqlpass,$mysqlhost,$cfg->{'mysqldumpopts'},$mysqldb,$targetname);
    }
  system($cmdline);
  if ($? == -1)
    {
    print(STDERR "ExportMySQL(mysqldump) FAILED!!!\nReason: $!\n\n");
    WriteToLog(sprintf("ERROR: MySQL Export failed: %s (%s)",$!,$cfg->{'name'}[$setnr]));
    unlink($targetname);
    return(undef);
    }
  elsif ($? & 127)
    {
    printf(STDERR "ExportMySQL(mysqldump) FAILED!!!\n%s died with signal %d, %s coredump\n\n",$cfg->{'mysqldump'},($? & 127),  ($? & 128) ? 'with' : 'without');
    WriteToLog(sprintf("ERROR: MySQL Export died with signal %d, %s coredump (%s)",($? & 127),  ($? & 128) ? 'with' : 'without',$cfg->{'name'}[$setnr]));
    unlink($targetname);
    return(undef);
    }
  else
    {
    if(($? >> 8) > 0)
      {
      printf(STDERR "ExportMySQL(mysqldump) FAILED!!!\n%s exited with value %d\n\n",$cfg->{'mysqldump'}, $? >> 8);
      WriteToLog(sprintf("ERROR: MySQL Export exited with value %d (%s)",$? >> 8,$cfg->{'name'}[$setnr]));
      unlink($targetname);
      return(undef);
      }
    }
  # And GZIP the file (as we have only 1 file we skip the tar here):

  $cmdline = sprintf("%s %s >/dev/null 2>&1",$cfg->{'gzip'},$targetname);
  system($cmdline);
  if ($? == -1)
    {
    print(STDERR "ExportMySQL(gzip) FAILED!!!\nReason: $!\n\n");
    WriteToLog(sprintf("ERROR: MySQLExp->%s failed: %s (%s)",$cfg->{'gzip'},$!,$cfg->{'name'}[$setnr]));
    unlink($targetname);
    return(undef);
    }
  elsif ($? & 127)
    {
    printf(STDERR "ExportMySQL(gzip) FAILED!!!\n%s died with signal %d, %s coredump\n\n",$cfg->{'gzip'},($? & 127),  ($? & 128) ? 'with' : 'without');
    WriteToLog(sprintf("ERROR: MySQLExp->%s died with signal %d, %s coredump (%s)",$cfg->{'gzip'},($? & 127),  ($? & 128) ? 'with' : 'without',$cfg->{'name'}[$setnr]));
    unlink($targetname);
    return(undef);
    }
  else
    {
    if(($? >> 8) > 0)
      {
      printf(STDERR "ExportMySQL(gzip) FAILED!!!\n%s exited with value %d\n\n",$cfg->{'gzip'}, $? >> 8);
      WriteToLog(sprintf("ERROR: MySQLExp->%s exited with value %d (%s)",$cfg->{'gzip'},$? >> 8,$cfg->{'name'}[$setnr]));
      unlink($targetname.$cfg->{'ext'});
      unlink($targetname);
      return(undef);
      }
    }
  my $et = getmicrotime;
  printf("done (%2.3fs).\n",abs($et-$st));
  WriteToLog(sprintf("MySQL Export finished after %2.3fs",abs($et-$st)));
  return($targetname.$cfg->{'ext'});
  }

###################################################################################################
#    NAME: CheckGenerations()
# PURPOSE: Performs check on available generations for a given backupset. If more generations
#          available than configured this routine removes the oldest entries.
#   INPUT: $setnr     => Backupset number to work on
#          $calltype  => How we are called: 0 = Commandline (interactive) | 1 = internal
#  RETURN: None
###################################################################################################

sub CheckGenerations($$)
  {
  my ($setnr,$calltype) = @_;
  my $target      = '';
  my $filecount   = 0;
  my $currentmode = 0;
  my @availfiles;
  my $removedfiles= 0;
  my @ftpkillfiles;       # We copy here all ftp files to remove and remove them at once (if any)

  if(!defined($cfg->{'maxgen'}[$setnr]) || int($cfg->{'maxgen'}[$setnr])==0)    # We only check if MAXGEN is configured AND > 0
    {
    return;
    }
  # V0.38: If given setnr is a pattern, skip it.
  if($cfg->{'pattern'}[$setnr] == 1)
    {
    return;
    }
  my @test = split(/,/,$cfg->{'mode'}[$setnr]);
  for(my $a =0; $a < scalar @test; $a++)
    {
    if($test[$a] eq 'fcopy')
      {
      $target = "FCOPY (".$cfg->{'dest'}[$setnr].")";
      @availfiles = ReadFileCopies($setnr, $cfg->{'dest'}[$setnr]);
      $currentmode = MODE_FCOPY;
      }
    elsif($test[$a] eq 'ftp')
      {
      my @dummy = split(/\@/,$cfg->{'ftp'}[$setnr]);
      $target = "FTP (".$dummy[1].")";
      @availfiles = ReadFTPCopies($setnr);
      $currentmode = MODE_FTP;
      }
    else
      {
      die("Unknown mode ".$test[$a]." detected!!!!");
      }
    # V0.38: In case of a problem we return here instead of trying to work on a null array...
    if(!defined($availfiles[0]) || scalar @availfiles == 0)
      {
      return;
      }
    @availfiles = sort(@availfiles);
    $filecount  = scalar @availfiles;
    if(!$calltype)
      {
      printf("Set #%d uses target = %s with %d files (maximum to save: %d backups)\n",$setnr,$target,scalar @availfiles,$cfg->{'maxgen'}[$setnr]);
      }
    if($filecount > $cfg->{'maxgen'}[$setnr])
      {
      for(my $r = 0; $r < $filecount - $cfg->{'maxgen'}[$setnr]; $r++)
        {
        if($currentmode == MODE_FCOPY)
          {
          my $delname = sprintf("%s/%s",$cfg->{'dest'}[$setnr],$availfiles[$r]);
          if(!$calltype)
            {
            printf("Removing file %s\n",$delname);
            }
          WriteToLog(sprintf("CheckGen(): Removing file %s",$delname));
          unlink $delname;
          $removedfiles++;
          }
        elsif($currentmode == MODE_FTP)
          {
          push @ftpkillfiles, $availfiles[$r];
          }
        }
      if(@ftpkillfiles)
        {
        $removedfiles+=RemoveFTPFiles($setnr,@ftpkillfiles);
        undef @ftpkillfiles;
        }
      } # End filecount > ..
    }
  if($removedfiles)
    {
    printf("Check Generations: Removed %d backupfile(s).\n\n",$removedfiles);
    WriteToLog(sprintf("Removed %d backupfile(s)",$removedfiles));
    }
  }

###################################################################################################
#    NAME: RemoveFTPFiles()
# PURPOSE: Removes a list of files from $setnr ftp host
#   INPUT: $setnr   => Backupset number to work on
#          @rmfiles => Array of files to remove from given $setnr FTPHOST.
#  RETURN: Amount of files removed
###################################################################################################

sub RemoveFTPFiles($@)
  {
  my ($setnr, @rmfiles) = @_;
  my $rfiles  = 0;
  my $ftp;

  $cfg->{'ftp'}[$setnr]=~/(.*)(\:)(.*)(@)(.*)/;
  my $ftpuser = $1;
  my $ftppass = $3;
  my $ftphost = $5;
  my @hostdata = split(/\:/,$ftphost);

  # V0.38: If given setnr is a pattern, skip it.
  if($cfg->{'pattern'}[$setnr] == 1)
    {
    return(0);
    }
  if(!defined($hostdata[1]))
    {
    $hostdata[1] = '21';  # Default port is 21
    }
  eval
    {
    $ftp = Net::FTP->new($hostdata[0], Debug => 0, Port => $hostdata[1], Timeout => FTP_TIMEOUT) or die "$@";
    };
  if($@)
    {
    $@=~/(.*)(at.*)/i;
    print(STDERR "RemoveFTPFiles(connect) FAILED: Reason: $1\n\n");
    WriteToLog(sprintf("ERROR: RemoveFTPFiles(#%d) FAILED: %s",$setnr,$1));
    return(undef);
    }
  if(!$ftp->login($ftpuser,$ftppass))
    {
    print(STDERR "RemoveFTPFiles(login) FAILED!\nFTP-Server says: ".$ftp->message."\n");
    WriteToLog(sprintf("ERROR: RemoveFTPFiles(#%d) LOGIN FAILED: %s",$setnr,$ftp->message));
    $ftp->quit;
    return(undef);
    }
  if($cfg->{'ftpdir'}[$setnr] ne '')
    {
    if(!$ftp->cwd($cfg->{'ftpdir'}[$setnr]))
      {
      print(STDERR "RemoveFTPFiles(cwd) failed!\nFTP-Server says: ".$ftp->message."\n");
      WriteToLog(sprintf("ERROR: RemoveFTPFiles(#%d) CWD FAILED: %s",$setnr,$ftp->message));
      $ftp->quit;
      return(undef);
      }
    }
  for(my $i = 0; $i < scalar @rmfiles; $i++)
    {
    eval
      {
      $ftp->delete($rmfiles[$i]);
      WriteToLog(sprintf("CheckGen(): Removing FTP file %s",$rmfiles[$i]));
      $rfiles++;
      };
    if($@)  # In case of an error we inform the user but continue.
      {
      printf(STDERR "RemoveFTPFiles(delete) failed!\nFTP-Server says: %s\n",$ftp->message);
      WriteToLog(sprintf("ERROR: RemoveFTPFiles(%d)->delete(%s) failed: %s\n",$setnr,$rmfiles[$i],$ftp->message));
      }
    }
  $ftp->quit;
  return($rfiles);
  }

###################################################################################################
#    NAME: ReadFileCopies()
# PURPOSE: Reads available backups from filesystem and returns array of found backup filenames
#   INPUT: $setnr   => Backupset number to work on
#          $target  => Parsed target directory where to find the files.
#  RETURN: Array of found backups or undef in case of an error
###################################################################################################

sub ReadFileCopies($$)
  {
  my ($setnr, $targetdir) = @_;
  my @backupfiles;
  my @entries;
  my $namematch = $cfg->{'name'}[$setnr].'_\d{8}_\d{6}';    # RegEx to find the backupfiles

  # V0.38: If given setnr is a pattern, skip it.
  if($cfg->{'pattern'}[$setnr] == 1)
    {
    return(0);
    }
  opendir THISDIR, $targetdir || die "cannot open $targetdir for reading\n";
  my @files = readdir THISDIR;
  close THISDIR;
  @entries = grep !/^\.\.?$/,@files;
  @entries = sort @entries;
  for(my $i = 0; $i < scalar @entries; $i++)
    {
    if($entries[$i]=~/^$namematch/)
      {
      push @backupfiles, $entries[$i];
      }
    }
  return(@backupfiles);
  }

###################################################################################################
#    NAME: ReadFTPCopies()
# PURPOSE: Reads available backups from FTP server and returns array of found backup filenames
#   INPUT: $setnr   => Backupset number to work on
#  RETURN: Array of found backups or undef in case of an error
###################################################################################################

sub ReadFTPCopies($)
  {
  my ($setnr) = @_;
  my $namematch = $cfg->{'name'}[$setnr].'_\d{8}_\d{6}';    # RegEx to find the backupfiles
  my @entries;
  my @backupfiles;
  my $ftp;

  # V0.38: If given setnr is a pattern, skip it.
  if($cfg->{'pattern'}[$setnr] == 1)
    {
    return(0);
    }
  $cfg->{'ftp'}[$setnr]=~/(.*)(\:)(.*)(@)(.*)/;
  my $ftpuser = $1;
  my $ftppass = $3;
  my $ftphost = $5;
  my @hostdata = split(/\:/,$ftphost);
  if(!defined($hostdata[1]))
    {
    $hostdata[1] = '21';  # Default port is 21
    }
  eval
    {
    $ftp = Net::FTP->new($hostdata[0], Debug => 0, Port => $hostdata[1], Timeout => FTP_TIMEOUT) or die "$@";
    };
  if($@)
    {
    $@=~/(.*)(at.*)/i;
    print(STDERR "ReadFTPCopies(connect) FAILED: Reason: $1\n\n");
    WriteToLog(sprintf("ERROR: ReadFTPCopies(#%d) FAILED: %s",$setnr,$1));
    return(undef);
    }
  if(!$ftp->login($ftpuser,$ftppass))
    {
    print(STDERR "ReadFTPCopies(login) FAILED!\nFTP-Server says: ".$ftp->message."\n");
    WriteToLog(sprintf("ERROR: ReadFTPCopies(#%d) LOGIN FAILED: %s",$setnr,$ftp->message));
    $ftp->quit;
    return(undef);
    }
  if($cfg->{'ftpdir'}[$setnr] ne '')
    {
    if(!$ftp->cwd($cfg->{'ftpdir'}[$setnr]))
      {
      print(STDERR "ReadFTPCopies(cwd) failed!\nFTP-Server says: ".$ftp->message."\n");
      WriteToLog(sprintf("ERROR: ReadFTPCopies(#%d) CWD FAILED: %s",$setnr,$ftp->message));
      $ftp->quit;
      return(undef);
      }
    }
  @entries = $ftp->ls;
  $ftp->quit;
  # Now filter out the data for the current backupset name:
  for(my $i = 0; $i < scalar @entries; $i++)
    {
    if($entries[$i]=~/^$namematch/)
      {
      push @backupfiles, $entries[$i];
      }
    }
  return(@backupfiles);
  }

###################################################################################################
#    NAME: ReadMySQLDBList()
# PURPOSE: Reads available databases from given MySQL DB, matches them against the pattern and
#          returns an array of database names which will be added to our internal config list
#   INPUT: $muser => Username for MySQL, must have access to the "SHOW DATABASES" command!
#          $mpass => Password of the given user
#          $mhost => Hostname to connect
#          $mdb   => Pattern to use against the list of available databases
#          $mdump => Full path to mysqldump util from configuration
#  RETURN: Array of found databases or undef in case of an error
#   NOTES: We assume here that the MYSQL tool "mysql" can be found under the same location
#          as the MySQLDump command. if this is not the case we abort here with an appropiate
#          error message.
#          mysql --user=<user> --password=<pass> --host=<host> --vertical --execute="show databases"
####################################################################################################

sub ReadMySQLDBList($$$$$$)
  {
  my ($muser,$mpass,$mhost,$mdb,$mdump,$tmpdir) = @_;
  my $commandline = '';
  my @resultarray;

  if(!defined($mdump) || $mdump eq '')
    {
    die("ERROR: No MySQLDump utility configured - MySQL support is disabled!!!\n");
    }
  my $mtool = dirname($mdump)."/mysql";
  if(!-X $mtool)
    {
    die(sprintf("ERROR: mysql utility under %s is not executeable!!!\n",$mtool));
    }
  my $tmpname = sprintf("%s/dblist_%s.lst",$tmpdir,$$);
  if($mpass eq '')
    {
    $commandline = sprintf("%s --user=%s --host=%s --vertical --execute=\"show databases\" >%s",$mtool,$muser,$mhost,$tmpname);
    }
  else
    {
    $commandline = sprintf("%s --user=%s --password=%s --host=%s --vertical --execute=\"show databases\" >%s",$mtool,$muser,$mpass,$mhost,$tmpname);
    }
  system($commandline);
  if ($? == -1)
    {
    print(STDERR "ReadMySQLDBList() FAILED!!!\nReason: $!\n\n");
    WriteToLog(sprintf("ERROR: ReadMySQLDBList() failed: %s",$!));
    unlink($tmpname);
    return(undef);
    }
  elsif ($? & 127)
    {
    printf(STDERR "ReadMySQLDBList() FAILED!!!\n%s died with signal %d, %s coredump\n\n",($? & 127),  ($? & 128) ? 'with' : 'without');
    WriteToLog(sprintf("ERROR: ReadMySQLDBList() died with signal %d, %s coredump",($? & 127),  ($? & 128) ? 'with' : 'without'));
    unlink($tmpname);
    return(undef);
    }
  else
    {
    if(($? >> 8) > 0)
      {
      printf(STDERR "ReadMySQLDBList() FAILED!!!\n%s exited with value %d\n\n", $? >> 8);
      WriteToLog(sprintf("ERROR: ReadMySQLDBList() exited with value %d",$? >> 8));
      unlink($tmpname);
      return(undef);
      }
    }
  # Now read and parse the temporary file extracting the MySQL Database names:
  open(FH, $tmpname) || die "ReadMySQLDBList(): Unable to read temporary database list file?!";
  while(<FH>)
    {
    if($_ =~/^database.*/i)
      {
      my $data = $_;
      $data=~/(^database\:\s)(.*)/i;
      my $dbname = $2;
      push @resultarray,$dbname;
      }
    }
  close(FH);
  unlink($tmpname);
  my @entries = grep /^$mdb/,@resultarray;
  return(sort @entries);
  }

###################################################################################################

__END__

=head1 NAME

SGLBackup.pl - Configurable backupper for various projects and/or Databases.

=head1 SYNOPSIS

SGLBackup.pl [options]

 Options:
   --help       Brief help message
   --man        full documentation
   --config     Alternative configfile (default is config.ini)
   --backupset  Number of backupset to process (default is all)
   --showconfig Lists parsed configuration and exit
   --checkgen   Performs checks on available generations

=head1 OPTIONS

=over 14

=item B<--help>

Print a brief help message and exits.

=item B<--man>

Prints the manual page and exits.

=item B<--config>

Specify an alternative configuration file.
If you ommit this parameter the default of
B<config.ini> is used.

=item B<--backupset>

Pass here the backupset number you wish to process.
If this parameter is not set all backupsets are processed.

=item B<--showconfig>

Dumps out the parsed configuration and exits.

=item B<--checkgen>

Checks amount of available backup files on destination and
clean up oldest files if more than configured files are found.
This process is also performed after every backup, here you can
call it manually if you wish. Note that only these backup files
are checked which have the MAXGEN_<x> parameter set, all other
are skipped.

=back

=head1 DESCRIPTION

This script will take a number of directories, tars one by one
and compress it with your favourite packer. Finally the packed
files can be sent via FTP, copied to an backup device etc.
Additional to normal filesystem backups one can backup also
Oracle and MySQL Databases with this script. To have Database
support you need "mysqldump" for MySQL and a full installation
of Oracle (either server or client) to have Oracle support enabled.

=head1 AUTHOR

Written and (c) 2003-2006 by Sascha 'SieGeL' Pfalz <webmaster@saschapfalz.de>

Released under the GNU public licence.

=head1 VERSION

This is SGLBackup.pl B<V0.39>

For changes see file CHANGELOG.

=cut
